@@ -25,10 +25,11 @@ package Astro::FITS::Header;
 #    Brad Cavanagh (b.cavanagh@jach.hawaii.edu)
 
 #  Revision:
-#     $Id: Header.pm,v 3.0 2006/08/19 00:09:06 timj Exp $
+#     $Id$
 
 #  Copyright:
-#     Copyright (C) 2001-2005 Particle Physics and Astronomy Research Council.
+#     Copyright (C) 2001-2006 Particle Physics and Astronomy Research Council.
+#     Copyright (C) 2007-2009 Science and Technology Facilities Council.
 #     Portions copyright (C) 2002 Southwest Research Institute
 #     All Rights Reserved.
 
@@ -60,7 +61,7 @@ use Carp;
 
 use Astro::FITS::Header::Item;
 
-$VERSION = sprintf("%d.%03d", q$Revision: 3.0 $ =~ /(\d+)\.(\d+)/);
+$VERSION = 3.01;
 
 # Operator overloads
 use overload '""' => "stringify",
@@ -70,7 +71,7 @@ use overload '""' => "stringify",
 
 =head1 REVISION
 
-$Id: Header.pm,v 3.0 2006/08/19 00:09:06 timj Exp $
+$Id$
 
 =head1 METHODS
 
@@ -97,10 +98,10 @@ sub new {
   # bless the header block into the class
   my $block = bless { HEADER => [],
                       LOOKUP  => {},
-		      LASTKEY => undef,
-		      TieRetRef => 0,
+                      LASTKEY => undef,
+                      TieRetRef => 0,
                       SUBHDRS => [],
-		    }, $class;
+                    }, $class;
 
   # Configure the object, even with no arguments since configure
   # still puts the minimum SIMPLE card in.
@@ -160,25 +161,25 @@ in the header using an C<Astro::FITS::Header::Item> of type 'HEADER'.
 =cut
 
 sub subhdrs {
-    my $self = shift;
+  my $self = shift;
 
-    if (@_) {
-        # verify the class
-        my $i;
-        for my $h (@_) {
-	  croak "Argument $i supplied to subhdrs method is not a Astro::FITS::Header object\n"
-	    unless UNIVERSAL::isa( $h, "Astro::FITS::Header" );
-          $i++;
-        }
+  if (@_) {
+    # verify the class
+    my $i;
+    for my $h (@_) {
+      croak "Argument $i supplied to subhdrs method is not a Astro::FITS::Header object\n"
+        unless UNIVERSAL::isa( $h, "Astro::FITS::Header" );
+      $i++;
+    }
 
-        # store them
-        @{$self->{SUBHDRS}} = @_;
-    }
-    if (wantarray()) {
-        return @{$self->{SUBHDRS}};
-    } else {
-        return $self->{SUBHDRS};
-    }
+    # store them
+    @{$self->{SUBHDRS}} = @_;
+  }
+  if (wantarray()) {
+    return @{$self->{SUBHDRS}};
+  } else {
+    return $self->{SUBHDRS};
+  }
 }
 
 =item B<item>
@@ -191,13 +192,13 @@ does not exist.
 =cut
 
 sub item {
-   my ( $self, $index ) = @_;
+  my ( $self, $index ) = @_;
 
-   return undef unless defined $index;
-   return undef unless exists ${$self->{HEADER}}[$index];
+  return undef unless defined $index;
+  return undef unless exists ${$self->{HEADER}}[$index];
 
-   # grab and return the Header::Item at $index
-   return ${$self->{HEADER}}[$index];
+  # grab and return the Header::Item at $index
+  return ${$self->{HEADER}}[$index];
 }
 
 
@@ -211,15 +212,15 @@ FITS Header.
 =cut
 
 sub get_wcs {
-   my $self = shift;
+  my $self = shift;
 
-   require Starlink::AST;
-   my $fchan = Starlink::AST::FitsChan->new();
-   for my $i ( $self->cards() ) {
-      $fchan->PutFits( $i, 0);
-   }
-   $fchan->Clear( "Card" );
-   return $fchan->Read();
+  require Starlink::AST;
+  my $fchan = Starlink::AST::FitsChan->new();
+  for my $i ( $self->cards() ) {
+    $fchan->PutFits( $i, 0);
+  }
+  $fchan->Clear( "Card" );
+  return $fchan->Read();
    
 }
 
@@ -235,13 +236,13 @@ Returns keyword referenced by index, C<undef> if it does not exist.
 =cut
 
 sub keyword {
-   my ( $self, $index ) = @_;
+  my ( $self, $index ) = @_;
 
-   return undef unless defined $index;
-   return undef unless exists ${$self->{HEADER}}[$index];
+  return undef unless defined $index;
+  return undef unless exists ${$self->{HEADER}}[$index];
 
-   # grab and return the keyword at $index
-   return ${$self->{HEADER}}[$index]->keyword();
+  # grab and return the keyword at $index
+  return ${$self->{HEADER}}[$index]->keyword();
 }
 
 # I T E M   B Y   N A M E  -------------------------------------------------
@@ -261,11 +262,11 @@ may be a regular expression created with the C<qr> operator.
 =cut
 
 sub itembyname {
-   my ( $self, $keyword ) = @_;
+  my ( $self, $keyword ) = @_;
 
-   my @items = @{$self->{HEADER}}[$self->index($keyword)];
+  my @items = @{$self->{HEADER}}[$self->index($keyword)];
 
-   return wantarray ?  @items : @items ? $items[0] : undef;
+  return wantarray ?  @items : @items ? $items[0] : undef;
 
 }
 
@@ -284,16 +285,16 @@ context. See C<Astro::FITS::Header::Item> for a list of allowed types.
 =cut
 
 sub itembytype {
-   my ( $self, $type ) = @_;
+  my ( $self, $type ) = @_;
 
-   return () unless defined $type;
+  return () unless defined $type;
 
-   $type = uc($type);
+  $type = uc($type);
 
-   # No optimised lookup so brute force it
-   my @items = grep { $_->type eq $type } @{ $self->{HEADER} };
+  # No optimised lookup so brute force it
+  my @items = grep { $_->type eq $type } @{ $self->{HEADER} };
 
-   return wantarray ?  @items : @items ? $items[0] : undef;
+  return wantarray ?  @items : @items ? $items[0] : undef;
 
 }
 
@@ -315,27 +316,24 @@ C<undef> if the keyword does not exist.
 =cut
 
 sub index {
-   my ( $self, $keyword ) = @_;
+  my ( $self, $keyword ) = @_;
    
-   # grab the index array from lookup table
-   my @index;
+  # grab the index array from lookup table
+  my @index;
 
-   if ( 'Regexp' eq ref $keyword )
-   {
-     push @index, @{$self->{LOOKUP}{$_}}
-       foreach grep { /$keyword/ && 
-			defined $self->{LOOKUP}{$_} } keys %{$self->{LOOKUP}};
-     @index = sort @index;
-   }
-   else
-   {
-     @index = @{${$self->{LOOKUP}}{$keyword}}
-       if ( exists ${$self->{LOOKUP}}{$keyword} && 
-	    defined ${$self->{LOOKUP}}{$keyword} );
-   }
+  if ( 'Regexp' eq ref $keyword ) {
+    push @index, @{$self->{LOOKUP}{$_}}
+      foreach grep { /$keyword/ && 
+                       defined $self->{LOOKUP}{$_} } keys %{$self->{LOOKUP}};
+    @index = sort @index;
+  } else {
+    @index = @{${$self->{LOOKUP}}{$keyword}}
+      if ( exists ${$self->{LOOKUP}}{$keyword} && 
+           defined ${$self->{LOOKUP}}{$keyword} );
+  }
    
-   # return the values array
-   return wantarray ? @index : @index ? $index[0] : undef;
+  # return the values array
+  return wantarray ? @index : @index ? $index[0] : undef;
 
 }
 
@@ -355,13 +353,13 @@ C<undef> if the keyword does not exist.
 =cut
 
 sub value {
-   my ( $self, $keyword ) = @_;
+  my ( $self, $keyword ) = @_;
 
-   # resolve the values from the index array from lookup table
-   my @values = map { ${$self->{HEADER}}[$_]->value() } $self->index($keyword);
+  # resolve the values from the index array from lookup table
+  my @values = map { ${$self->{HEADER}}[$_]->value() } $self->index($keyword);
 
-   # loop over the indices and grab the values
-   return wantarray ? @values : @values ? $values[0] : undef;
+  # loop over the indices and grab the values
+  return wantarray ? @values : @values ? $values[0] : undef;
 
 }
 
@@ -383,14 +381,14 @@ C<undef> if the keyword does not exist.
 =cut
 
 sub comment {
-   my ( $self, $keyword ) = @_;
+  my ( $self, $keyword ) = @_;
 
-   # resolve the comments from the index array from lookup table
-   my @comments =
-     map { ${$self->{HEADER}}[$_]->comment() } $self->index($keyword);
+  # resolve the comments from the index array from lookup table
+  my @comments =
+    map { ${$self->{HEADER}}[$_]->comment() } $self->index($keyword);
 
-   # loop over the indices and grab the comments
-   return wantarray ?  @comments : @comments ? $comments[0] : undef;
+  # loop over the indices and grab the comments
+  return wantarray ?  @comments : @comments ? $comments[0] : undef;
 }
 
 # I N S E R T -------------------------------------------------------------
@@ -410,13 +408,13 @@ The insert position can be negative.
 =cut
 
 sub insert{
-   my ($self, $index, $item) = @_;
+  my ($self, $index, $item) = @_;
 
-   # splice the new FITS header card into the array
-   # Splice automatically triggers a lookup table rebuild
-   $self->splice($index, 0, $item);
+  # splice the new FITS header card into the array
+  # Splice automatically triggers a lookup table rebuild
+  $self->splice($index, 0, $item);
 
-   return;
+  return;
 }
 
 
@@ -433,10 +431,10 @@ returns the replaced card.
 =cut
 
 sub replace{
-   my ($self, $index, $item) = @_;
-   # remove the specified item and replace with $item
-   # Splice triggers a rebuild so we do not have to
-   return $self->splice( $index, 1, $item);
+  my ($self, $index, $item) = @_;
+  # remove the specified item and replace with $item
+  # Splice triggers a rebuild so we do not have to
+  return $self->splice( $index, 1, $item);
 }
 
 # R E M O V E -------------------------------------------------------------
@@ -452,10 +450,10 @@ returns the removed card.
 =cut
 
 sub remove{
-   my ($self, $index) = @_;
-   # remove the  FITS header card from the array
-   # Splice always triggers a lookup table rebuild so we don't have to
-   return $self->splice( $index, 1);
+  my ($self, $index) = @_;
+  # remove the  FITS header card from the array
+  # Splice always triggers a lookup table rebuild so we don't have to
+  return $self->splice( $index, 1);
 }
 
 # R E P L A C E  B Y  N A M E ---------------------------------------------
@@ -472,21 +470,21 @@ created with the C<qr> operator.
 =cut
 
 sub replacebyname{
-   my ($self, $keyword, $item) = @_;
+  my ($self, $keyword, $item) = @_;
 
-   # grab the index array from lookup table
-   my @index = $self->index($keyword);
+  # grab the index array from lookup table
+  my @index = $self->index($keyword);
 
-   # loop over the keywords
-   # We use a real splice rather than the class splice for efficiency
-   # in order to prevent an index rebuild for each index
-   my @cards = map { splice @{$self->{HEADER}}, $_, 1, $item;} @index;
+  # loop over the keywords
+  # We use a real splice rather than the class splice for efficiency
+  # in order to prevent an index rebuild for each index
+  my @cards = map { splice @{$self->{HEADER}}, $_, 1, $item;} @index;
 
-   # force rebuild
-   $self->_rebuild_lookup;
+  # force rebuild
+  $self->_rebuild_lookup;
 
-   # return removed items
-   return wantarray ? @cards : $cards[scalar(@cards)-1];
+  # return removed items
+  return wantarray ? @cards : $cards[scalar(@cards)-1];
 
 }
 
@@ -504,21 +502,24 @@ created with the C<qr> operator.
 =cut
 
 sub removebyname{
-   my ($self, $keyword) = @_;
+  my ($self, $keyword) = @_;
 
-   # grab the index array from lookup table
-   my @index = $self->index($keyword);
+  # grab the index array from lookup table
+  my @index = $self->index($keyword);
 
-   # loop over the keywords
-   # We use a real splice rather than the class splice for efficiency
-   # in order to prevent an index rebuild for each index
-   my @cards = map { splice @{$self->{HEADER}}, $_, 1; } @index;
+  # loop over the keywords
+  # We use a real splice rather than the class splice for efficiency
+  # in order to prevent an index rebuild for each index. The ugly code
+  # is needed in case we have multiple indices returned, which can
+  # happen if we have a regular expression passed in as a keyword.
+  my $i = -1;
+  my @cards = map { $i++; splice @{$self->{HEADER}}, ( $_ - $i ), 1; } sort @index;
 
-   # force rebuild
-   $self->_rebuild_lookup;
+  # force rebuild
+  $self->_rebuild_lookup;
 
-   # return removed items
-   return wantarray ? @cards : $cards[scalar(@cards)-1];
+  # return removed items
+  return wantarray ? @cards : $cards[scalar(@cards)-1];
 }
 
 # S P L I C E --------------------------------------------------------------
@@ -538,45 +539,47 @@ is negative, counts from the end of the FITS header.
 =cut
 
 sub splice {
-   my $self = shift;
-   my ($offset, $length, @list) = @_;
+  my $self = shift;
+  my ($offset, $length, @list) = @_;
 
-   # If the array is empty and we get a negative offset we
-   # must convert it to an offset of 0 to prevent a:
-   #   Modification of non-creatable array value attempted, subscript -1
-   # fatal error
-   # This can occur with a tied hash and the %{$tieref} = %new
-   # construct
-   if (defined $offset) {
-     $offset = 0 if (@{$self->{HEADER}} == 0 && $offset < 0);
-   }
+  # If the array is empty and we get a negative offset we
+  # must convert it to an offset of 0 to prevent a:
+  #   Modification of non-creatable array value attempted, subscript -1
+  # fatal error
+  # This can occur with a tied hash and the %{$tieref} = %new
+  # construct
+  if (defined $offset) {
+    $offset = 0 if (@{$self->{HEADER}} == 0 && $offset < 0);
+  }
 
-   # the removed cards
-   my @cards;
+  # the removed cards
+  my @cards;
 
-   if (@list) {
-     # all arguments supplied
-     for my $i (@list) {
-       croak "Arguments to splice must be Astro::FITS::Header::Item objects"
-         unless UNIVERSAL::isa($i, "Astro::FITS::Header::Item");	
-     }
-     @cards = splice @{$self->{HEADER}}, $offset, $length, @list;
+  if (@list) {
+    # all arguments supplied
+    my $n = 0;
+    for my $i (@list) {
+      croak "Argument $n to splice must be Astro::FITS::Header::Item objects"
+        unless UNIVERSAL::isa($i, "Astro::FITS::Header::Item");	
+      $n++;
+    }
+    @cards = splice @{$self->{HEADER}}, $offset, $length, @list;
 
-   } elsif (defined $length) {
-     # length and (presumably) offset
-     @cards = splice @{$self->{HEADER}}, $offset, $length;
+  } elsif (defined $length) {
+    # length and (presumably) offset
+    @cards = splice @{$self->{HEADER}}, $offset, $length;
 
-   } elsif (defined $offset) {
-     # offset only
-     @cards = splice @{$self->{HEADER}}, $offset;
-   } else {
-     # none
-     @cards = splice @{$self->{HEADER}};
-   }
+  } elsif (defined $offset) {
+    # offset only
+    @cards = splice @{$self->{HEADER}}, $offset;
+  } else {
+    # none
+    @cards = splice @{$self->{HEADER}};
+  }
 
-   # update the internal lookup table and return
-   $self->_rebuild_lookup();
-   return wantarray ? @cards : $cards[scalar(@cards)-1];
+  # update the internal lookup table and return
+  $self->_rebuild_lookup();
+  return wantarray ? @cards : $cards[scalar(@cards)-1];
 }
 
 # C A R D S --------------------------------------------------------------
@@ -619,8 +622,8 @@ Returns the header as an array of FITS::Header:Item objects.
 =cut
 
 sub allitems {
-   my $self = shift;
-   return map { $_ } @{$self->{HEADER}};
+  my $self = shift;
+  return map { $_ } @{$self->{HEADER}};
 }
 
 # C O N F I G U R E -------------------------------------------------------
@@ -633,48 +636,83 @@ sub allitems {
 
 =item B<configure>
 
-Configures the object, takes an array of FITS header cards 
-or an array of Astro::FITS::Header::Item objects as input.
+Configures the object, takes an array of FITS header cards, 
+an array of Astro::FITS::Header::Item objects or a simple hash as input.
 If you feed in nothing at all, it uses a default array containing
 just the SIMPLE card required at the top of all FITS files.
 
   $header->configure( Cards => \@array );
   $header->configure( Items => \@array );
+  $header->configure( Hash => \%hash );
 
-Does nothing if the array is not supplied.
+Does nothing if the array is not supplied. If the hash scheme is used
+and the hash contains the special key of SUBHEADERS pointing to an
+array of hashes, these will be read as proper sub headers. All other
+references in the hash will be ignored. Note that the default key
+order will be retained in the object created via the hash.
 
 =cut
 
 sub configure {
-    my $self = shift;
+  my $self = shift;
     
-    # grab the argument list
-    my %args = @_;
+  # grab the argument list
+  my %args = @_;
     
-    if (exists $args{Cards} && defined $args{Cards}) {
+  if (exists $args{Cards} && defined $args{Cards}) {
 	
-	# First translate each incoming card into a Item object
-	# Any existing cards are removed
-	@{$self->{HEADER}} = map {
+    # First translate each incoming card into a Item object
+    # Any existing cards are removed
+    @{$self->{HEADER}} = map {
 	    new Astro::FITS::Header::Item( Card => $_ );
-	} @{ $args{Cards} };
+    } @{ $args{Cards} };
 	
     # Now build the lookup table. There would be a slight efficiency
-	# gain to include this in a loop over the cards but prefer
-	# to reuse the method for this rather than repeating code
-	$self->_rebuild_lookup;
+    # gain to include this in a loop over the cards but prefer
+    # to reuse the method for this rather than repeating code
+    $self->_rebuild_lookup;
 	
-    } elsif (exists $args{Items} && defined $args{Items}){
-	# We have an array of Astro::FITS::Header::Items
-	@{$self->{HEADER}} = @{ $args{Items} };
-	$self->_rebuild_lookup;
-    } elsif( !defined($self->{HEADER}) ||  !@{$self->{HEADER}} ) {
-	@{$self->{HEADER}} = (
-	      new Astro::FITS::Header::Item( Card=> "SIMPLE  =  T"),
-	      new Astro::FITS::Header::Item( Card=> "END", Type=>"END" )
-			      );
-	$self->_rebuild_lookup; 
+  } elsif (exists $args{Items} && defined $args{Items}) {
+    # We have an array of Astro::FITS::Header::Items
+    @{$self->{HEADER}} = @{ $args{Items} };
+    $self->_rebuild_lookup;
+  } elsif (exists $args{Hash} && defined $args{Hash} ) {
+    # we have a hash so convert to Item objects and store
+    # use a For loop instead of map since we want to
+    # skip some items
+    croak "Hash constructor requested but not given a hash reference"
+      unless ref($args{Hash}) eq 'HASH';
+    my @items;
+    my @subheaders;
+    for my $k (keys %{$args{Hash}}) {
+      if ($k eq 'SUBHEADERS' 
+          && ref($args{Hash}->{$k}) eq 'ARRAY'
+          && ref($args{Hash}->{$k}->[0]) eq 'HASH') {
+        # special case
+        @subheaders = map { $self->new( Hash => $_ ) } @{$args{Hash}->{$k}};
+      } elsif (not ref($args{Hash}->{$k})) {
+        # if we have new lines in the value, we should duplicate the item
+        # so split on new lines
+        my $value = $args{Hash}->{$k};
+        $value = '' unless defined $value;
+        my @lines = split(/^/m,$value);
+        chomp(@lines);          # remove the newlines
+
+        push(@items, map { new Astro::FITS::Header::Item( Keyword => $k,
+                                                          Value => $_ ) }
+             @lines);
+      }
     }
+    @{$self->{HEADER}} = @items;
+    $self->_rebuild_lookup;
+    $self->subhdrs(@subheaders) if @subheaders;
+  } elsif ( !defined($self->{HEADER}) ||  !@{$self->{HEADER}} ) {
+    @{$self->{HEADER}} = (
+                          new Astro::FITS::Header::Item( Card=> "SIMPLE  =  T"),
+                          new Astro::FITS::Header::Item( Card=> "END", Type=>"END" )
+                         );
+    $self->_rebuild_lookup; 
+  }
 }
 
 =item B<merge_primary>
@@ -717,8 +755,8 @@ sub merge_primary {
 
   # optional options handling
   my %opt = ( merge_unique => 0,
-	      force_return_diffs => 0,
-	    );
+              force_return_diffs => 0,
+            );
   if (ref($_[0]) eq 'HASH') {
     my $o = shift;
     %opt = ( %opt, %$o );
@@ -749,21 +787,24 @@ sub merge_primary {
     for my $item ($hdr->allitems) {
       my $key;
       my $type = $item->type;
-      if ($type eq 'COMMENT' || $type eq 'UNDEF') {
-	$key = $item->card;
+      if (!defined $type || $type eq 'BLANK') {
+        # blank line so skip it
+        next;
+      } elsif ($type eq 'COMMENT' || $type eq 'UNDEF') {
+        $key = $item->card;
       } elsif ($type eq 'HEADER') {
-	next;
+        next;
       } else {
-	$key = $item->keyword;
+        $key = $item->keyword;
       }
 
       if (exists $items{$key}) {
-	# Store the item, but in a hash with key corresponding
-	# to the input header number
-	push( @{ $items{$key}}, { item => $item, hnum => $hnum } );
+        # Store the item, but in a hash with key corresponding
+        # to the input header number
+        push( @{ $items{$key}}, { item => $item, hnum => $hnum } );
       } else {
-	$items{$key} = [ { item => $item, hnum => $hnum } ];
-	push(@order, $key);
+        $items{$key} = [ { item => $item, hnum => $hnum } ];
+        push(@order, $key);
       }
     }
     $hnum++;
@@ -786,7 +827,7 @@ sub merge_primary {
     for my $i (@items[1..$#items]) {
       # Ask the Items to compare using the equals() method
       if ($items[0]->{item}->equals( $i->{item} )) {
-	$match++;
+        $match++;
       }
     }
 
@@ -803,7 +844,7 @@ sub merge_primary {
       # Not enough of the items matched. Store to the relevant difference
       # arrays.
       for my $i (@items) {
-	push(@{ $difference[$i->{hnum}] }, $i->{item});
+        push(@{ $difference[$i->{hnum}] }, $i->{item});
       }
 
     }
@@ -819,10 +860,10 @@ sub merge_primary {
   # for returning it
   unshift(@difference, \@merged );
 
-  # convert back to FITS object, making sure we stringify the Item
-  # objects so that we retain copies
+  # convert back to FITS object, Construct using the Items directly
+  # - they will be copied without strinfication.
   for my $d (@difference) {
-    $d = $self->new( Cards => [ map { "$_" } @$d ]);
+    $d = $self->new( Cards => $d );
   }
 
   # remembering that the merged array is on the front
@@ -841,6 +882,62 @@ sub freeze {
   return bless $self, 'Astro::FITS::Header';
 }
 
+=item B<append>
+
+Append or update a card.
+
+  $header->append( $card );
+
+This method can take either an Astro::FITS::Header::Item object, an
+Astro::FITS::Header object, or a reference to an array of
+Astro::FITS::Header::Item objects.
+
+In all cases, if the given Astro::FITS::Header::Item keyword exists in
+the header, then the value will be overwritten with the one passed to
+the method. Otherwise, the card will be appended to the end of the
+header.
+
+Nothing is returned.
+
+=cut
+
+sub append {
+  my $self = shift;
+  my $thing = shift;
+
+  my @cards;
+  if ( UNIVERSAL::isa( $thing, "Astro::FITS::Header::Item" ) ) {
+    push @cards, $thing;
+  } elsif ( UNIVERSAL::isa( $thing, "Astro::FITS::Header" ) ) {
+    @cards = $thing->allitems;
+  } elsif ( ref( $thing ) eq 'ARRAY' ) {
+    @cards = @$thing;
+  }
+
+  foreach my $card ( @cards ) {
+    my $item = $self->itembyname( $card->keyword );
+    if ( defined( $item ) ) {
+
+      # Update the given card.
+      $self->replacebyname( $card->keyword, $card )
+
+    } else {
+
+      # Don't append a SIMPLE header as that can lead to disaster and
+      # strife and gnashing of teeth (and violates the FITS standard).
+      next if ( uc( $card->keyword ) eq 'SIMPLE' );
+
+      # Retrieve the index of the END card, and insert this card
+      # before that one, but only if the END header actually exists.
+      my $index = $self->index( 'END' );
+      $index = ( defined( $index ) ? $index : -1 );
+      $self->insert( $index, $card );
+    }
+  }
+
+  $self->_rebuild_lookup;
+}
+
 # P R I V A T  E   M E T H O D S ------------------------------------------
 
 =back
@@ -880,29 +977,29 @@ to the indices of all header cards following the modifed card.
 =cut
 
 sub _rebuild_lookup {
-   my $self = shift;
+  my $self = shift;
    
-   # rebuild the lookup table
+  # rebuild the lookup table
 
-   # empty the hash 
-   $self->{LOOKUP} = { };
+  # empty the hash 
+  $self->{LOOKUP} = { };
 
-   # loop over the existing header array
-   for my $j (0 .. $#{$self->{HEADER}}) {
+  # loop over the existing header array
+  for my $j (0 .. $#{$self->{HEADER}}) {
 
-      # grab the keyword from each header item;
-      my $key = ${$self->{HEADER}}[$j]->keyword();
+    # grab the keyword from each header item;
+    my $key = ${$self->{HEADER}}[$j]->keyword();
             
-      # need to account to repeated keywords (e.g. COMMENT)
-      unless ( exists ${$self->{LOOKUP}}{$key} &&
-               defined ${$self->{LOOKUP}}{$key} ) {
-         # new keyword
-         ${$self->{LOOKUP}}{$key} = [ $j ];
-      } else {     
-         # keyword exists, push the current index into the array
-         push( @{${$self->{LOOKUP}}{$key}}, $j );
-      }   
-   }
+    # need to account to repeated keywords (e.g. COMMENT)
+    unless ( exists ${$self->{LOOKUP}}{$key} &&
+             defined ${$self->{LOOKUP}}{$key} ) {
+      # new keyword
+      ${$self->{LOOKUP}}{$key} = [ $j ];
+    } else {     
+      # keyword exists, push the current index into the array
+      push( @{${$self->{LOOKUP}}{$key}}, $j );
+    }   
+  }
 
 }
 
@@ -1085,9 +1182,9 @@ interface.
 
 # List of known comment-type fields
 %Astro::FITS::Header::COMMENT_FIELD = (
-  "COMMENT"=>1,
-  "HISTORY"=>1
-);
+                                       "COMMENT"=>1,
+                                       "HISTORY"=>1
+                                      );
 
 
 # constructor
@@ -1136,7 +1233,7 @@ sub FETCH {
   # if they're asking for $key_COMMENT.
   my $item;
   my $t_ok;
-  if( $wantcomment || $key =~ /^COMMENT$/ || $key =~ /^END$/) {
+  if ( $wantcomment || $key =~ /^(COMMENT)|(HISTORY)$/ || $key =~ /^END$/) {
     $item = ($self->itembyname($key))[0];
     $t_ok = (defined $item) && (defined $item->type);
     $wantvalue = 0 if ($t_ok && ($item->type eq 'COMMENT'));
@@ -1145,8 +1242,8 @@ sub FETCH {
   # The END card is a special case.  We always return " " for the value,
   # and undef for the comment.
   return ($wantvalue ? " " : undef)
-    if( ($t_ok && ($item->type eq 'END')) ||
-        ((defined $item) && ($key eq 'END')) );
+    if ( ($t_ok && ($item->type eq 'END')) ||
+         ((defined $item) && ($key eq 'END')) );
 
   # Retrieve all the values/comments. Note that we go through the entire
   # header for this in case of multiple matches
@@ -1207,9 +1304,9 @@ sub FETCH {
   # first element anyway.)
   my $out;
   if ($self->tiereturnsref && scalar(@out) > 1) {
-      $out = \@out;
+    $out = \@out;
   } else {
-      $out = $out[0];
+    $out = $out[0];
   }
   
   return $out;
@@ -1244,9 +1341,9 @@ sub STORE {
   # Make sure that we do not stringify reference arguments by mistake
   # when looking from slashes
 
-  if ( !ref($value) && $keyword !~ m/(COMMENT)|(HISTORY)/ and 
-      $value =~ s:\s*(?<!\\)/\s*(.*)::         # Identify any '/' not preceded by '\'
-      ) { 
+  if (defined $value && !ref($value) && $keyword !~ m/(COMMENT$)|(^HISTORY$)/ and
+      $value =~ s:\s*(?<!\\)/\s*(.*):: # Identify any '/' not preceded by '\'
+     ) { 
     my $comment = $1;
 
     # Recurse to store the comment.  This is a direct (non-method) call to 
@@ -1256,14 +1353,17 @@ sub STORE {
   }
 
   # unescape (unless we are blessed)
-  if (!ref($value)) {
+  if (defined $value && !ref($value)) {
     $value =~ s:\\\\:\\:g;
     $value =~ s:\\\/:\/:g;
   }
 
   # skip the shenanigans for the normal case
   # or if we have an Astro::FITS::Header
-  if (UNIVERSAL::isa($value, "Astro::FITS::Header")) {
+  if (!defined $value) {
+    @values = ($value);
+
+  } elsif (UNIVERSAL::isa($value, "Astro::FITS::Header")) {
     @values = ($value);
 
   } elsif (ref $value eq 'HASH') {
@@ -1275,16 +1375,10 @@ sub STORE {
       @values = ($tied);
     } else {
       # Convert it to a hash
-      my @items = map { new Astro::FITS::Header::Item( Keyword => $_,
-                                                       Value => $value->{$_}
-                                                     ) } keys (%{$value});
-
-      # Create the Header object.
-      @values = (new Astro::FITS::Header( Cards => \@items ));
-
+      @values = ( Astro::FITS::Header->new( Hash => $value ) );
     }
 
-  } elsif((ref $value eq 'ARRAY') || (length $value > 70) || $value =~ m/\n/s ) {
+  } elsif ((ref $value eq 'ARRAY') || (length $value > 70) || $value =~ m/\n/s ) {
     my @val;
     # @val gets intermediate breakdowns, @values gets line-by-line breakdowns.
 
@@ -1295,7 +1389,7 @@ sub STORE {
     } elsif (ref $value) {
       croak "Can't put non-array ref values into a tied FITS header\n";
 
-    } elsif( $value =~ m/\n/s ) {
+    } elsif ( $value =~ m/\n/s ) {
       @val = split("\n",$value);
       chomp @val;
 
@@ -1306,13 +1400,13 @@ sub STORE {
     # Cut up really long items into multiline strings
     my($val);
     foreach $val(@val) {
-      while((length $val) > 70) {
-	push(@values,substr($val,0,69)."\\");
-	$val = substr($val,69);
+      while ((length $val) > 70) {
+        push(@values,substr($val,0,69)."\\");
+        $val = substr($val,69);
       }
       push(@values,$val);
     }
-  }   ## End of complicated case
+  }                             ## End of complicated case
   else {
 
 
@@ -1323,43 +1417,43 @@ sub STORE {
   # Upper case the relevant item name
   $keyword = uc($keyword);
   
-  if($keyword eq 'END') {
-      # Special case for END keyword
-      # (drops value on floor, makes sure there is one END at the end)
-      my @index = $self->index($keyword);
-      if( @index != 1   ||   $index[0] != $#{$self->allitems}) {
-	  my $i;
-	  while(defined($i = shift @index)) {
+  if ($keyword eq 'END') {
+    # Special case for END keyword
+    # (drops value on floor, makes sure there is one END at the end)
+    my @index = $self->index($keyword);
+    if ( @index != 1   ||   $index[0] != $#{$self->allitems}) {
+      my $i;
+      while (defined($i = shift @index)) {
 	      $self->remove($i);
-	  }
       }
-      unless( @index ) {
-	  my $endcard = new Astro::FITS::Header::Item(Keyword=>'END',
-						      Type=>'END', 
-						      Value=>1);
-	  $self->insert( scalar ($self->allitems) , $endcard );
-      }
-      return;
+    }
+    unless( @index ) {
+      my $endcard = new Astro::FITS::Header::Item(Keyword=>'END',
+                                                  Type=>'END', 
+                                                  Value=>1);
+      $self->insert( scalar ($self->allitems) , $endcard );
+    }
+    return;
       
   } 
   
-  if($keyword eq 'SIMPLE') {
-      # Special case for SIMPLE keyword
-      # (sets value correctly, makes sure there is one SIMPLE at the beginning)
-      my @index = $self->index($keyword);
-      if( @index != 1  ||  $index[0] != 0) {
-	  my $i;
-	  while(defined ($i=shift @index)) {
+  if ($keyword eq 'SIMPLE') {
+    # Special case for SIMPLE keyword
+    # (sets value correctly, makes sure there is one SIMPLE at the beginning)
+    my @index = $self->index($keyword);
+    if ( @index != 1  ||  $index[0] != 0) {
+      my $i;
+      while (defined ($i=shift @index)) {
 	      $self->remove($i);
-	  }
       }
-      unless( @index ) {
-	  my $simplecard = new Astro::FITS::Header::Item(Keyword=>'SIMPLE',
-							 Value=>$values[0],
-							 Type=>'LOGICAL');
-	  $self->insert(0, $simplecard);
-      }
-      return;
+    }
+    unless( @index ) {
+      my $simplecard = new Astro::FITS::Header::Item(Keyword=>'SIMPLE',
+                                                     Value=>$values[0],
+                                                     Type=>'LOGICAL');
+      $self->insert(0, $simplecard);
+    }
+    return;
   }
   
 
@@ -1373,7 +1467,7 @@ sub STORE {
   my @items = $self->itembyname($keyword);
 
   ## Remove extra items if necessary
-  if(scalar(@items) > scalar(@values)) {
+  if (scalar(@items) > scalar(@values)) {
     my(@indices) = $self->index($keyword);
     my($i);
     for $i (1..(scalar(@items) - scalar(@values))) {
@@ -1382,7 +1476,7 @@ sub STORE {
   }
 
   ## Allocate new items if necessary
-  while(scalar(@items) < scalar(@values)) {
+  while (scalar(@items) < scalar(@values)) {
 
     my $item = new Astro::FITS::Header::Item(Keyword=>$keyword,Value=>undef);
     # (No need to set type here; Item does it for us)
@@ -1394,7 +1488,7 @@ sub STORE {
   ## Set values or comments
   my($i); 
   for $i(0..$#values) {
-    if($Astro::FITS::Header::COMMENT_FIELD{$keyword}) {
+    if ($Astro::FITS::Header::COMMENT_FIELD{$keyword}) {
       $items[$i]->type('COMMENT');
       $items[$i]->comment($values[$i]);
     } elsif (! $havevalue) {
@@ -1420,12 +1514,17 @@ sub EXISTS {
     return ( scalar(@{$self->subhdrs}) > 0 ? 1 : 0);
   }
 
-  if( !exists( ${$self->{LOOKUP}}{$keyword} ) ) {
+  if ( !exists( ${$self->{LOOKUP}}{$keyword} ) ) {
     return undef;
   }
-  if(  exists( ${$self->{LOOKUP}}{$keyword} ) &&
-      ${$self->{HEADER}}[${$self->{LOOKUP}}{$keyword}[0]]->type eq 'COMMENT' ) {
-    return undef;
+
+  # if we are being asked for a keyword that is associated with a COMMENT or BLANK
+  # type we return FALSE for existence. An undef type means we have to assume a valid
+  # item with unknown type
+  if (  exists( ${$self->{LOOKUP}}{$keyword} ) ) {
+    my $item = ${$self->{HEADER}}[${$self->{LOOKUP}}{$keyword}[0]];
+    my $type = $item->type;
+    return undef if (defined $type && ($type eq 'COMMENT' || $type eq 'BLANK') );
   }
 
   return 1;
@@ -1462,21 +1561,20 @@ sub NEXTKEY {
 
   # abort if the number of keys we have served equals the number in the
   # header array. One wrinkle is that if we have SUBHDRS we want to go
-  # rount one more time
+  # round one more time
+  
   if ($self->{LASTKEY}+1 == scalar(@{$self->{HEADER}})) {
-    if (scalar(@{ $self->subhdrs}) && !$self->{SEENKEY}->{SUBHEADERS}) {
-      $self->{SEENKEY}->{SUBHEADERS} = 1;
-      return "SUBHEADERS";
-    }
-    return undef
+    return $self->_check_for_subhdr();
   }
 
-  # Skip later lines of multi-line cards...
+  # Skip later lines of multi-line cards since the tie interface
+  # will return all the lines for a single keyword request.
   my($a);
   do {
     $self->{LASTKEY} += 1;  
     $a = $self->{HEADER}->[$self->{LASTKEY}];
-    return undef unless defined $a;
+    # Got to end of header if we do not have $a
+    return $self->_check_for_subhdr() unless defined $a;
   } while ( $self->{SEENKEY}->{$a->keyword});
   $a = $a->keyword;
 
@@ -1484,6 +1582,18 @@ sub NEXTKEY {
   return $a;
 }
 
+# called if we have run out of normal keys
+#  args: $self Returns: undef or "SUBHEADER"
+sub _check_for_subhdr {
+  my $self = shift;
+  if (scalar(@{ $self->subhdrs}) && !$self->{SEENKEY}->{SUBHEADERS}) {
+    $self->{SEENKEY}->{SUBHEADERS} = 1;
+    return "SUBHEADERS";
+  }
+  return undef;
+}
+
+
 # garbage collection
 # sub DESTROY { }
 
@@ -1631,7 +1741,6 @@ sub _hdr_to_tie {
     tie %header, ref($hdr), $hdr;
     return \%header;
   }
-  print "UNDEF\n";
   return undef;
 }
 
